import sys
from typing import Dict, List, Optional

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from localstack.aws.api import RequestContext, ServiceException, ServiceRequest, handler

Boolean = bool
BoxedInteger = int
Integer = int
MessageAttributeName = str
String = str
TagKey = str
TagValue = str
Token = str


class MessageSystemAttributeName(str):
    SenderId = "SenderId"
    SentTimestamp = "SentTimestamp"
    ApproximateReceiveCount = "ApproximateReceiveCount"
    ApproximateFirstReceiveTimestamp = "ApproximateFirstReceiveTimestamp"
    SequenceNumber = "SequenceNumber"
    MessageDeduplicationId = "MessageDeduplicationId"
    MessageGroupId = "MessageGroupId"
    AWSTraceHeader = "AWSTraceHeader"


class MessageSystemAttributeNameForSends(str):
    AWSTraceHeader = "AWSTraceHeader"


class QueueAttributeName(str):
    All = "All"
    Policy = "Policy"
    VisibilityTimeout = "VisibilityTimeout"
    MaximumMessageSize = "MaximumMessageSize"
    MessageRetentionPeriod = "MessageRetentionPeriod"
    ApproximateNumberOfMessages = "ApproximateNumberOfMessages"
    ApproximateNumberOfMessagesNotVisible = "ApproximateNumberOfMessagesNotVisible"
    CreatedTimestamp = "CreatedTimestamp"
    LastModifiedTimestamp = "LastModifiedTimestamp"
    QueueArn = "QueueArn"
    ApproximateNumberOfMessagesDelayed = "ApproximateNumberOfMessagesDelayed"
    DelaySeconds = "DelaySeconds"
    ReceiveMessageWaitTimeSeconds = "ReceiveMessageWaitTimeSeconds"
    RedrivePolicy = "RedrivePolicy"
    FifoQueue = "FifoQueue"
    ContentBasedDeduplication = "ContentBasedDeduplication"
    KmsMasterKeyId = "KmsMasterKeyId"
    KmsDataKeyReusePeriodSeconds = "KmsDataKeyReusePeriodSeconds"
    DeduplicationScope = "DeduplicationScope"
    FifoThroughputLimit = "FifoThroughputLimit"
    RedriveAllowPolicy = "RedriveAllowPolicy"
    SqsManagedSseEnabled = "SqsManagedSseEnabled"


class BatchEntryIdsNotDistinct(ServiceException):
    """Two or more batch entries in the request have the same ``Id``."""

    pass


class BatchRequestTooLong(ServiceException):
    """The length of all the messages put together is more than the limit."""

    pass


class EmptyBatchRequest(ServiceException):
    """The batch request doesn't contain any entries."""

    pass


class InvalidAttributeName(ServiceException):
    """The specified attribute doesn't exist."""

    pass


class InvalidBatchEntryId(ServiceException):
    """The ``Id`` of a batch entry in a batch request doesn't abide by the
    specification.
    """

    pass


class InvalidIdFormat(ServiceException):
    """The specified receipt handle isn't valid for the current version."""

    pass


class InvalidMessageContents(ServiceException):
    """The message contains characters outside the allowed set."""

    pass


class MessageNotInflight(ServiceException):
    """The specified message isn't in flight."""

    pass


class OverLimit(ServiceException):
    """The specified action violates a limit. For example, ``ReceiveMessage``
    returns this error if the maximum number of inflight messages is reached
    and ``AddPermission`` returns this error if the maximum number of
    permissions for the queue is reached.
    """

    pass


class PurgeQueueInProgress(ServiceException):
    """Indicates that the specified queue previously received a ``PurgeQueue``
    request within the last 60 seconds (the time it can take to delete the
    messages in the queue).
    """

    pass


class QueueDeletedRecently(ServiceException):
    """You must wait 60 seconds after deleting a queue before you can create
    another queue with the same name.
    """

    pass


class QueueDoesNotExist(ServiceException):
    """The specified queue doesn't exist."""

    pass


class QueueNameExists(ServiceException):
    """A queue with this name already exists. Amazon SQS returns this error
    only if the request includes attributes whose values differ from those
    of the existing queue.
    """

    pass


class ReceiptHandleIsInvalid(ServiceException):
    """The specified receipt handle isn't valid."""

    pass


class TooManyEntriesInBatchRequest(ServiceException):
    """The batch request contains more entries than permissible."""

    pass


class UnsupportedOperation(ServiceException):
    """Error code 400. Unsupported operation."""

    pass


AWSAccountIdList = List[String]
ActionNameList = List[String]


class AddPermissionRequest(ServiceRequest):
    QueueUrl: String
    Label: String
    AWSAccountIds: AWSAccountIdList
    Actions: ActionNameList


AttributeNameList = List[QueueAttributeName]


class BatchResultErrorEntry(TypedDict, total=False):
    """Gives a detailed description of the result of an action on each entry in
    the request.
    """

    Id: String
    SenderFault: Boolean
    Code: String
    Message: Optional[String]


BatchResultErrorEntryList = List[BatchResultErrorEntry]
Binary = bytes
BinaryList = List[Binary]


class ChangeMessageVisibilityBatchRequestEntry(TypedDict, total=False):
    """Encloses a receipt handle and an entry id for each message in
    ``ChangeMessageVisibilityBatch.``

    All of the following list parameters must be prefixed with
    ``ChangeMessageVisibilityBatchRequestEntry.n``, where ``n`` is an
    integer value starting with ``1``. For example, a parameter list for
    this action might look like this:

    ``&ChangeMessageVisibilityBatchRequestEntry.1.Id=change_visibility_msg_2``

    ``&ChangeMessageVisibilityBatchRequestEntry.1.ReceiptHandle=your_receipt_handle``

    ``&ChangeMessageVisibilityBatchRequestEntry.1.VisibilityTimeout=45``
    """

    Id: String
    ReceiptHandle: String
    VisibilityTimeout: Optional[Integer]


ChangeMessageVisibilityBatchRequestEntryList = List[ChangeMessageVisibilityBatchRequestEntry]


class ChangeMessageVisibilityBatchRequest(ServiceRequest):
    QueueUrl: String
    Entries: ChangeMessageVisibilityBatchRequestEntryList


class ChangeMessageVisibilityBatchResultEntry(TypedDict, total=False):
    """Encloses the ``Id`` of an entry in ``ChangeMessageVisibilityBatch.``"""

    Id: String


ChangeMessageVisibilityBatchResultEntryList = List[ChangeMessageVisibilityBatchResultEntry]


class ChangeMessageVisibilityBatchResult(TypedDict, total=False):
    """For each message in the batch, the response contains a
    ``ChangeMessageVisibilityBatchResultEntry`` tag if the message succeeds
    or a ``BatchResultErrorEntry`` tag if the message fails.
    """

    Successful: ChangeMessageVisibilityBatchResultEntryList
    Failed: BatchResultErrorEntryList


class ChangeMessageVisibilityRequest(ServiceRequest):
    QueueUrl: String
    ReceiptHandle: String
    VisibilityTimeout: Integer


TagMap = Dict[TagKey, TagValue]
QueueAttributeMap = Dict[QueueAttributeName, String]


class CreateQueueRequest(ServiceRequest):
    QueueName: String
    Attributes: Optional[QueueAttributeMap]
    tags: Optional[TagMap]


class CreateQueueResult(TypedDict, total=False):
    """Returns the ``QueueUrl`` attribute of the created queue."""

    QueueUrl: Optional[String]


class DeleteMessageBatchRequestEntry(TypedDict, total=False):
    """Encloses a receipt handle and an identifier for it."""

    Id: String
    ReceiptHandle: String


DeleteMessageBatchRequestEntryList = List[DeleteMessageBatchRequestEntry]


class DeleteMessageBatchRequest(ServiceRequest):
    QueueUrl: String
    Entries: DeleteMessageBatchRequestEntryList


class DeleteMessageBatchResultEntry(TypedDict, total=False):
    """Encloses the ``Id`` of an entry in ``DeleteMessageBatch.``"""

    Id: String


DeleteMessageBatchResultEntryList = List[DeleteMessageBatchResultEntry]


class DeleteMessageBatchResult(TypedDict, total=False):
    """For each message in the batch, the response contains a
    ``DeleteMessageBatchResultEntry`` tag if the message is deleted or a
    ``BatchResultErrorEntry`` tag if the message can't be deleted.
    """

    Successful: DeleteMessageBatchResultEntryList
    Failed: BatchResultErrorEntryList


class DeleteMessageRequest(ServiceRequest):
    QueueUrl: String
    ReceiptHandle: String


class DeleteQueueRequest(ServiceRequest):
    QueueUrl: String


class GetQueueAttributesRequest(ServiceRequest):
    QueueUrl: String
    AttributeNames: Optional[AttributeNameList]


class GetQueueAttributesResult(TypedDict, total=False):
    """A list of returned queue attributes."""

    Attributes: Optional[QueueAttributeMap]


class GetQueueUrlRequest(ServiceRequest):
    QueueName: String
    QueueOwnerAWSAccountId: Optional[String]


class GetQueueUrlResult(TypedDict, total=False):
    """For more information, see `Interpreting
    Responses <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-api-responses.html>`__
    in the *Amazon SQS Developer Guide*.
    """

    QueueUrl: Optional[String]


class ListDeadLetterSourceQueuesRequest(ServiceRequest):
    QueueUrl: String
    NextToken: Optional[Token]
    MaxResults: Optional[BoxedInteger]


QueueUrlList = List[String]


class ListDeadLetterSourceQueuesResult(TypedDict, total=False):
    """A list of your dead letter source queues."""

    queueUrls: QueueUrlList
    NextToken: Optional[Token]


class ListQueueTagsRequest(ServiceRequest):
    QueueUrl: String


class ListQueueTagsResult(TypedDict, total=False):
    Tags: Optional[TagMap]


class ListQueuesRequest(ServiceRequest):
    QueueNamePrefix: Optional[String]
    NextToken: Optional[Token]
    MaxResults: Optional[BoxedInteger]


class ListQueuesResult(TypedDict, total=False):
    """A list of your queues."""

    QueueUrls: Optional[QueueUrlList]
    NextToken: Optional[Token]


StringList = List[String]


class MessageAttributeValue(TypedDict, total=False):
    """The user-specified message attribute value. For string data types, the
    ``Value`` attribute has the same restrictions on the content as the
    message body. For more information, see ``SendMessage.``

    ``Name``, ``type``, ``value`` and the message body must not be empty or
    null. All parts of the message attribute, including ``Name``, ``Type``,
    and ``Value``, are part of the message size restriction (256 KB or
    262,144 bytes).
    """

    StringValue: Optional[String]
    BinaryValue: Optional[Binary]
    StringListValues: Optional[StringList]
    BinaryListValues: Optional[BinaryList]
    DataType: String


MessageBodyAttributeMap = Dict[String, MessageAttributeValue]
MessageSystemAttributeMap = Dict[MessageSystemAttributeName, String]


class Message(TypedDict, total=False):
    """An Amazon SQS message."""

    MessageId: Optional[String]
    ReceiptHandle: Optional[String]
    MD5OfBody: Optional[String]
    Body: Optional[String]
    Attributes: Optional[MessageSystemAttributeMap]
    MD5OfMessageAttributes: Optional[String]
    MessageAttributes: Optional[MessageBodyAttributeMap]


MessageAttributeNameList = List[MessageAttributeName]


class MessageSystemAttributeValue(TypedDict, total=False):
    """The user-specified message system attribute value. For string data
    types, the ``Value`` attribute has the same restrictions on the content
    as the message body. For more information, see ``SendMessage.``

    ``Name``, ``type``, ``value`` and the message body must not be empty or
    null.
    """

    StringValue: Optional[String]
    BinaryValue: Optional[Binary]
    StringListValues: Optional[StringList]
    BinaryListValues: Optional[BinaryList]
    DataType: String


MessageBodySystemAttributeMap = Dict[
    MessageSystemAttributeNameForSends, MessageSystemAttributeValue
]
MessageList = List[Message]


class PurgeQueueRequest(ServiceRequest):
    QueueUrl: String


class ReceiveMessageRequest(ServiceRequest):
    QueueUrl: String
    AttributeNames: Optional[AttributeNameList]
    MessageAttributeNames: Optional[MessageAttributeNameList]
    MaxNumberOfMessages: Optional[Integer]
    VisibilityTimeout: Optional[Integer]
    WaitTimeSeconds: Optional[Integer]
    ReceiveRequestAttemptId: Optional[String]


class ReceiveMessageResult(TypedDict, total=False):
    """A list of received messages."""

    Messages: Optional[MessageList]


class RemovePermissionRequest(ServiceRequest):
    QueueUrl: String
    Label: String


class SendMessageBatchRequestEntry(TypedDict, total=False):
    """Contains the details of a single Amazon SQS message along with an
    ``Id``.
    """

    Id: String
    MessageBody: String
    DelaySeconds: Optional[Integer]
    MessageAttributes: Optional[MessageBodyAttributeMap]
    MessageSystemAttributes: Optional[MessageBodySystemAttributeMap]
    MessageDeduplicationId: Optional[String]
    MessageGroupId: Optional[String]


SendMessageBatchRequestEntryList = List[SendMessageBatchRequestEntry]


class SendMessageBatchRequest(ServiceRequest):
    QueueUrl: String
    Entries: SendMessageBatchRequestEntryList


class SendMessageBatchResultEntry(TypedDict, total=False):
    """Encloses a ``MessageId`` for a successfully-enqueued message in a
    ``SendMessageBatch.``
    """

    Id: String
    MessageId: String
    MD5OfMessageBody: String
    MD5OfMessageAttributes: Optional[String]
    MD5OfMessageSystemAttributes: Optional[String]
    SequenceNumber: Optional[String]


SendMessageBatchResultEntryList = List[SendMessageBatchResultEntry]


class SendMessageBatchResult(TypedDict, total=False):
    """For each message in the batch, the response contains a
    ``SendMessageBatchResultEntry`` tag if the message succeeds or a
    ``BatchResultErrorEntry`` tag if the message fails.
    """

    Successful: SendMessageBatchResultEntryList
    Failed: BatchResultErrorEntryList


class SendMessageRequest(ServiceRequest):
    QueueUrl: String
    MessageBody: String
    DelaySeconds: Optional[Integer]
    MessageAttributes: Optional[MessageBodyAttributeMap]
    MessageSystemAttributes: Optional[MessageBodySystemAttributeMap]
    MessageDeduplicationId: Optional[String]
    MessageGroupId: Optional[String]


class SendMessageResult(TypedDict, total=False):
    """The ``MD5OfMessageBody`` and ``MessageId`` elements."""

    MD5OfMessageBody: Optional[String]
    MD5OfMessageAttributes: Optional[String]
    MD5OfMessageSystemAttributes: Optional[String]
    MessageId: Optional[String]
    SequenceNumber: Optional[String]


class SetQueueAttributesRequest(ServiceRequest):
    QueueUrl: String
    Attributes: QueueAttributeMap


TagKeyList = List[TagKey]


class TagQueueRequest(ServiceRequest):
    QueueUrl: String
    Tags: TagMap


class UntagQueueRequest(ServiceRequest):
    QueueUrl: String
    TagKeys: TagKeyList


class SqsApi:

    service = "sqs"
    version = "2012-11-05"

    @handler("AddPermission")
    def add_permission(
        self,
        context: RequestContext,
        queue_url: String,
        label: String,
        aws_account_ids: AWSAccountIdList,
        actions: ActionNameList,
    ) -> None:
        """Adds a permission to a queue for a specific
        `principal <https://docs.aws.amazon.com/general/latest/gr/glos-chap.html#P>`__.
        This allows sharing access to the queue.

        When you create a queue, you have full control access rights for the
        queue. Only you, the owner of the queue, can grant or deny permissions
        to the queue. For more information about these permissions, see `Allow
        Developers to Write Messages to a Shared
        Queue <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-writing-an-sqs-policy.html#write-messages-to-shared-queue>`__
        in the *Amazon SQS Developer Guide*.

        -  ``AddPermission`` generates a policy for you. You can use
           ``SetQueueAttributes`` to upload your policy. For more information,
           see `Using Custom Policies with the Amazon SQS Access Policy
           Language <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-creating-custom-policies.html>`__
           in the *Amazon SQS Developer Guide*.

        -  An Amazon SQS policy can have a maximum of 7 actions.

        -  To remove the ability to change queue permissions, you must deny
           permission to the ``AddPermission``, ``RemovePermission``, and
           ``SetQueueAttributes`` actions in your IAM policy.

        Some actions take lists of parameters. These lists are specified using
        the ``param.n`` notation. Values of ``n`` are integers starting from 1.
        For example, a parameter list with two elements looks like this:

        ``&AttributeName.1=first``

        ``&AttributeName.2=second``

        Cross-account permissions don't apply to this action. For more
        information, see `Grant cross-account permissions to a role and a user
        name <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-customer-managed-policy-examples.html#grant-cross-account-permissions-to-role-and-user-name>`__
        in the *Amazon SQS Developer Guide*.

        :param queue_url: The URL of the Amazon SQS queue to which permissions are added.
        :param label: The unique identification of the permission you're setting (for example,
        ``AliceSendMessage``).
        :param aws_account_ids: The Amazon Web Services account numbers of the
        `principals <https://docs.
        :param actions: The action the client wants to allow for the specified principal.
        :raises OverLimit:
        """
        raise NotImplementedError

    @handler("ChangeMessageVisibility")
    def change_message_visibility(
        self,
        context: RequestContext,
        queue_url: String,
        receipt_handle: String,
        visibility_timeout: Integer,
    ) -> None:
        """Changes the visibility timeout of a specified message in a queue to a
        new value. The default visibility timeout for a message is 30 seconds.
        The minimum is 0 seconds. The maximum is 12 hours. For more information,
        see `Visibility
        Timeout <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-visibility-timeout.html>`__
        in the *Amazon SQS Developer Guide*.

        For example, you have a message with a visibility timeout of 5 minutes.
        After 3 minutes, you call ``ChangeMessageVisibility`` with a timeout of
        10 minutes. You can continue to call ``ChangeMessageVisibility`` to
        extend the visibility timeout to the maximum allowed time. If you try to
        extend the visibility timeout beyond the maximum, your request is
        rejected.

        An Amazon SQS message has three basic states:

        #. Sent to a queue by a producer.

        #. Received from the queue by a consumer.

        #. Deleted from the queue.

        A message is considered to be *stored* after it is sent to a queue by a
        producer, but not yet received from the queue by a consumer (that is,
        between states 1 and 2). There is no limit to the number of stored
        messages. A message is considered to be *in flight* after it is received
        from a queue by a consumer, but not yet deleted from the queue (that is,
        between states 2 and 3). There is a limit to the number of inflight
        messages.

        Limits that apply to inflight messages are unrelated to the *unlimited*
        number of stored messages.

        For most standard queues (depending on queue traffic and message
        backlog), there can be a maximum of approximately 120,000 inflight
        messages (received from a queue by a consumer, but not yet deleted from
        the queue). If you reach this limit, Amazon SQS returns the
        ``OverLimit`` error message. To avoid reaching the limit, you should
        delete messages from the queue after they're processed. You can also
        increase the number of queues you use to process your messages. To
        request a limit increase, `file a support
        request <https://console.aws.amazon.com/support/home#/case/create?issueType=service-limit-increase&limitType=service-code-sqs>`__.

        For FIFO queues, there can be a maximum of 20,000 inflight messages
        (received from a queue by a consumer, but not yet deleted from the
        queue). If you reach this limit, Amazon SQS returns no error messages.

        If you attempt to set the ``VisibilityTimeout`` to a value greater than
        the maximum time left, Amazon SQS returns an error. Amazon SQS doesn't
        automatically recalculate and increase the timeout to the maximum
        remaining time.

        Unlike with a queue, when you change the visibility timeout for a
        specific message the timeout value is applied immediately but isn't
        saved in memory for that message. If you don't delete a message after it
        is received, the visibility timeout for the message reverts to the
        original timeout value (not to the value you set using the
        ``ChangeMessageVisibility`` action) the next time the message is
        received.

        :param queue_url: The URL of the Amazon SQS queue whose message's visibility is changed.
        :param receipt_handle: The receipt handle associated with the message whose visibility timeout
        is changed.
        :param visibility_timeout: The new value for the message's visibility timeout (in seconds).
        :raises MessageNotInflight:
        :raises ReceiptHandleIsInvalid:
        """
        raise NotImplementedError

    @handler("ChangeMessageVisibilityBatch")
    def change_message_visibility_batch(
        self,
        context: RequestContext,
        queue_url: String,
        entries: ChangeMessageVisibilityBatchRequestEntryList,
    ) -> ChangeMessageVisibilityBatchResult:
        """Changes the visibility timeout of multiple messages. This is a batch
        version of ``ChangeMessageVisibility.`` The result of the action on each
        message is reported individually in the response. You can send up to 10
        ``ChangeMessageVisibility`` requests with each
        ``ChangeMessageVisibilityBatch`` action.

        Because the batch request can result in a combination of successful and
        unsuccessful actions, you should check for batch errors even when the
        call returns an HTTP status code of ``200``.

        Some actions take lists of parameters. These lists are specified using
        the ``param.n`` notation. Values of ``n`` are integers starting from 1.
        For example, a parameter list with two elements looks like this:

        ``&AttributeName.1=first``

        ``&AttributeName.2=second``

        :param queue_url: The URL of the Amazon SQS queue whose messages' visibility is changed.
        :param entries: A list of receipt handles of the messages for which the visibility
        timeout must be changed.
        :returns: ChangeMessageVisibilityBatchResult
        :raises TooManyEntriesInBatchRequest:
        :raises EmptyBatchRequest:
        :raises BatchEntryIdsNotDistinct:
        :raises InvalidBatchEntryId:
        """
        raise NotImplementedError

    @handler("CreateQueue")
    def create_queue(
        self,
        context: RequestContext,
        queue_name: String,
        attributes: QueueAttributeMap = None,
        tags: TagMap = None,
    ) -> CreateQueueResult:
        """Creates a new standard or FIFO queue. You can pass one or more
        attributes in the request. Keep the following in mind:

        -  If you don't specify the ``FifoQueue`` attribute, Amazon SQS creates
           a standard queue.

           You can't change the queue type after you create it and you can't
           convert an existing standard queue into a FIFO queue. You must either
           create a new FIFO queue for your application or delete your existing
           standard queue and recreate it as a FIFO queue. For more information,
           see `Moving From a Standard Queue to a FIFO
           Queue <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/FIFO-queues.html#FIFO-queues-moving>`__
           in the *Amazon SQS Developer Guide*.

        -  If you don't provide a value for an attribute, the queue is created
           with the default value for the attribute.

        -  If you delete a queue, you must wait at least 60 seconds before
           creating a queue with the same name.

        To successfully create a new queue, you must provide a queue name that
        adheres to the `limits related to
        queues <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/limits-queues.html>`__
        and is unique within the scope of your queues.

        After you create a queue, you must wait at least one second after the
        queue is created to be able to use the queue.

        To get the queue URL, use the ``GetQueueUrl`` action. ``GetQueueUrl``
        requires only the ``QueueName`` parameter. be aware of existing queue
        names:

        -  If you provide the name of an existing queue along with the exact
           names and values of all the queue's attributes, ``CreateQueue``
           returns the queue URL for the existing queue.

        -  If the queue name, attribute names, or attribute values don't match
           an existing queue, ``CreateQueue`` returns an error.

        Some actions take lists of parameters. These lists are specified using
        the ``param.n`` notation. Values of ``n`` are integers starting from 1.
        For example, a parameter list with two elements looks like this:

        ``&AttributeName.1=first``

        ``&AttributeName.2=second``

        Cross-account permissions don't apply to this action. For more
        information, see `Grant cross-account permissions to a role and a user
        name <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-customer-managed-policy-examples.html#grant-cross-account-permissions-to-role-and-user-name>`__
        in the *Amazon SQS Developer Guide*.

        :param queue_name: The name of the new queue.
        :param attributes: A map of attributes with their corresponding values.
        :param tags: Add cost allocation tags to the specified Amazon SQS queue.
        :returns: CreateQueueResult
        :raises QueueDeletedRecently:
        :raises QueueNameExists:
        """
        raise NotImplementedError

    @handler("DeleteMessage")
    def delete_message(
        self, context: RequestContext, queue_url: String, receipt_handle: String
    ) -> None:
        """Deletes the specified message from the specified queue. To select the
        message to delete, use the ``ReceiptHandle`` of the message (*not* the
        ``MessageId`` which you receive when you send the message). Amazon SQS
        can delete a message from a queue even if a visibility timeout setting
        causes the message to be locked by another consumer. Amazon SQS
        automatically deletes messages left in a queue longer than the retention
        period configured for the queue.

        The ``ReceiptHandle`` is associated with a *specific instance* of
        receiving a message. If you receive a message more than once, the
        ``ReceiptHandle`` is different each time you receive a message. When you
        use the ``DeleteMessage`` action, you must provide the most recently
        received ``ReceiptHandle`` for the message (otherwise, the request
        succeeds, but the message might not be deleted).

        For standard queues, it is possible to receive a message even after you
        delete it. This might happen on rare occasions if one of the servers
        which stores a copy of the message is unavailable when you send the
        request to delete the message. The copy remains on the server and might
        be returned to you during a subsequent receive request. You should
        ensure that your application is idempotent, so that receiving a message
        more than once does not cause issues.

        :param queue_url: The URL of the Amazon SQS queue from which messages are deleted.
        :param receipt_handle: The receipt handle associated with the message to delete.
        :raises InvalidIdFormat:
        :raises ReceiptHandleIsInvalid:
        """
        raise NotImplementedError

    @handler("DeleteMessageBatch")
    def delete_message_batch(
        self,
        context: RequestContext,
        queue_url: String,
        entries: DeleteMessageBatchRequestEntryList,
    ) -> DeleteMessageBatchResult:
        """Deletes up to ten messages from the specified queue. This is a batch
        version of ``DeleteMessage.`` The result of the action on each message
        is reported individually in the response.

        Because the batch request can result in a combination of successful and
        unsuccessful actions, you should check for batch errors even when the
        call returns an HTTP status code of ``200``.

        Some actions take lists of parameters. These lists are specified using
        the ``param.n`` notation. Values of ``n`` are integers starting from 1.
        For example, a parameter list with two elements looks like this:

        ``&AttributeName.1=first``

        ``&AttributeName.2=second``

        :param queue_url: The URL of the Amazon SQS queue from which messages are deleted.
        :param entries: A list of receipt handles for the messages to be deleted.
        :returns: DeleteMessageBatchResult
        :raises TooManyEntriesInBatchRequest:
        :raises EmptyBatchRequest:
        :raises BatchEntryIdsNotDistinct:
        :raises InvalidBatchEntryId:
        """
        raise NotImplementedError

    @handler("DeleteQueue")
    def delete_queue(self, context: RequestContext, queue_url: String) -> None:
        """Deletes the queue specified by the ``QueueUrl``, regardless of the
        queue's contents.

        Be careful with the ``DeleteQueue`` action: When you delete a queue, any
        messages in the queue are no longer available.

        When you delete a queue, the deletion process takes up to 60 seconds.
        Requests you send involving that queue during the 60 seconds might
        succeed. For example, a ``SendMessage`` request might succeed, but after
        60 seconds the queue and the message you sent no longer exist.

        When you delete a queue, you must wait at least 60 seconds before
        creating a queue with the same name.

        Cross-account permissions don't apply to this action. For more
        information, see `Grant cross-account permissions to a role and a user
        name <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-customer-managed-policy-examples.html#grant-cross-account-permissions-to-role-and-user-name>`__
        in the *Amazon SQS Developer Guide*.

        :param queue_url: The URL of the Amazon SQS queue to delete.
        """
        raise NotImplementedError

    @handler("GetQueueAttributes")
    def get_queue_attributes(
        self,
        context: RequestContext,
        queue_url: String,
        attribute_names: AttributeNameList = None,
    ) -> GetQueueAttributesResult:
        """Gets attributes for the specified queue.

        To determine whether a queue is
        `FIFO <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/FIFO-queues.html>`__,
        you can check whether ``QueueName`` ends with the ``.fifo`` suffix.

        :param queue_url: The URL of the Amazon SQS queue whose attribute information is
        retrieved.
        :param attribute_names: A list of attributes for which to retrieve information.
        :returns: GetQueueAttributesResult
        :raises InvalidAttributeName:
        """
        raise NotImplementedError

    @handler("GetQueueUrl")
    def get_queue_url(
        self,
        context: RequestContext,
        queue_name: String,
        queue_owner_aws_account_id: String = None,
    ) -> GetQueueUrlResult:
        """Returns the URL of an existing Amazon SQS queue.

        To access a queue that belongs to another AWS account, use the
        ``QueueOwnerAWSAccountId`` parameter to specify the account ID of the
        queue's owner. The queue's owner must grant you permission to access the
        queue. For more information about shared queue access, see
        ``AddPermission`` or see `Allow Developers to Write Messages to a Shared
        Queue <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-writing-an-sqs-policy.html#write-messages-to-shared-queue>`__
        in the *Amazon SQS Developer Guide*.

        :param queue_name: The name of the queue whose URL must be fetched.
        :param queue_owner_aws_account_id: The Amazon Web Services account ID of the account that created the
        queue.
        :returns: GetQueueUrlResult
        :raises QueueDoesNotExist:
        """
        raise NotImplementedError

    @handler("ListDeadLetterSourceQueues")
    def list_dead_letter_source_queues(
        self,
        context: RequestContext,
        queue_url: String,
        next_token: Token = None,
        max_results: BoxedInteger = None,
    ) -> ListDeadLetterSourceQueuesResult:
        """Returns a list of your queues that have the ``RedrivePolicy`` queue
        attribute configured with a dead-letter queue.

        The ``ListDeadLetterSourceQueues`` methods supports pagination. Set
        parameter ``MaxResults`` in the request to specify the maximum number of
        results to be returned in the response. If you do not set
        ``MaxResults``, the response includes a maximum of 1,000 results. If you
        set ``MaxResults`` and there are additional results to display, the
        response includes a value for ``NextToken``. Use ``NextToken`` as a
        parameter in your next request to ``ListDeadLetterSourceQueues`` to
        receive the next page of results.

        For more information about using dead-letter queues, see `Using Amazon
        SQS Dead-Letter
        Queues <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-dead-letter-queues.html>`__
        in the *Amazon SQS Developer Guide*.

        :param queue_url: The URL of a dead-letter queue.
        :param next_token: Pagination token to request the next set of results.
        :param max_results: Maximum number of results to include in the response.
        :returns: ListDeadLetterSourceQueuesResult
        :raises QueueDoesNotExist:
        """
        raise NotImplementedError

    @handler("ListQueueTags")
    def list_queue_tags(self, context: RequestContext, queue_url: String) -> ListQueueTagsResult:
        """List all cost allocation tags added to the specified Amazon SQS queue.
        For an overview, see `Tagging Your Amazon SQS
        Queues <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-queue-tags.html>`__
        in the *Amazon SQS Developer Guide*.

        Cross-account permissions don't apply to this action. For more
        information, see `Grant cross-account permissions to a role and a user
        name <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-customer-managed-policy-examples.html#grant-cross-account-permissions-to-role-and-user-name>`__
        in the *Amazon SQS Developer Guide*.

        :param queue_url: The URL of the queue.
        :returns: ListQueueTagsResult
        """
        raise NotImplementedError

    @handler("ListQueues")
    def list_queues(
        self,
        context: RequestContext,
        queue_name_prefix: String = None,
        next_token: Token = None,
        max_results: BoxedInteger = None,
    ) -> ListQueuesResult:
        """Returns a list of your queues in the current region. The response
        includes a maximum of 1,000 results. If you specify a value for the
        optional ``QueueNamePrefix`` parameter, only queues with a name that
        begins with the specified value are returned.

        The ``listQueues`` methods supports pagination. Set parameter
        ``MaxResults`` in the request to specify the maximum number of results
        to be returned in the response. If you do not set ``MaxResults``, the
        response includes a maximum of 1,000 results. If you set ``MaxResults``
        and there are additional results to display, the response includes a
        value for ``NextToken``. Use ``NextToken`` as a parameter in your next
        request to ``listQueues`` to receive the next page of results.

        Cross-account permissions don't apply to this action. For more
        information, see `Grant cross-account permissions to a role and a user
        name <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-customer-managed-policy-examples.html#grant-cross-account-permissions-to-role-and-user-name>`__
        in the *Amazon SQS Developer Guide*.

        :param queue_name_prefix: A string to use for filtering the list results.
        :param next_token: Pagination token to request the next set of results.
        :param max_results: Maximum number of results to include in the response.
        :returns: ListQueuesResult
        """
        raise NotImplementedError

    @handler("PurgeQueue")
    def purge_queue(self, context: RequestContext, queue_url: String) -> None:
        """Deletes the messages in a queue specified by the ``QueueURL`` parameter.

        When you use the ``PurgeQueue`` action, you can't retrieve any messages
        deleted from a queue.

        The message deletion process takes up to 60 seconds. We recommend
        waiting for 60 seconds regardless of your queue's size.

        Messages sent to the queue *before* you call ``PurgeQueue`` might be
        received but are deleted within the next minute.

        Messages sent to the queue *after* you call ``PurgeQueue`` might be
        deleted while the queue is being purged.

        :param queue_url: The URL of the queue from which the ``PurgeQueue`` action deletes
        messages.
        :raises QueueDoesNotExist:
        :raises PurgeQueueInProgress:
        """
        raise NotImplementedError

    @handler("ReceiveMessage")
    def receive_message(
        self,
        context: RequestContext,
        queue_url: String,
        attribute_names: AttributeNameList = None,
        message_attribute_names: MessageAttributeNameList = None,
        max_number_of_messages: Integer = None,
        visibility_timeout: Integer = None,
        wait_time_seconds: Integer = None,
        receive_request_attempt_id: String = None,
    ) -> ReceiveMessageResult:
        """Retrieves one or more messages (up to 10), from the specified queue.
        Using the ``WaitTimeSeconds`` parameter enables long-poll support. For
        more information, see `Amazon SQS Long
        Polling <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-long-polling.html>`__
        in the *Amazon SQS Developer Guide*.

        Short poll is the default behavior where a weighted random set of
        machines is sampled on a ``ReceiveMessage`` call. Thus, only the
        messages on the sampled machines are returned. If the number of messages
        in the queue is small (fewer than 1,000), you most likely get fewer
        messages than you requested per ``ReceiveMessage`` call. If the number
        of messages in the queue is extremely small, you might not receive any
        messages in a particular ``ReceiveMessage`` response. If this happens,
        repeat the request.

        For each message returned, the response includes the following:

        -  The message body.

        -  An MD5 digest of the message body. For information about MD5, see
           `RFC1321 <https://www.ietf.org/rfc/rfc1321.txt>`__.

        -  The ``MessageId`` you received when you sent the message to the
           queue.

        -  The receipt handle.

        -  The message attributes.

        -  An MD5 digest of the message attributes.

        The receipt handle is the identifier you must provide when deleting the
        message. For more information, see `Queue and Message
        Identifiers <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-queue-message-identifiers.html>`__
        in the *Amazon SQS Developer Guide*.

        You can provide the ``VisibilityTimeout`` parameter in your request. The
        parameter is applied to the messages that Amazon SQS returns in the
        response. If you don't include the parameter, the overall visibility
        timeout for the queue is used for the returned messages. For more
        information, see `Visibility
        Timeout <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-visibility-timeout.html>`__
        in the *Amazon SQS Developer Guide*.

        A message that isn't deleted or a message whose visibility isn't
        extended before the visibility timeout expires counts as a failed
        receive. Depending on the configuration of the queue, the message might
        be sent to the dead-letter queue.

        In the future, new attributes might be added. If you write code that
        calls this action, we recommend that you structure your code so that it
        can handle new attributes gracefully.

        :param queue_url: The URL of the Amazon SQS queue from which messages are received.
        :param attribute_names: A list of attributes that need to be returned along with each message.
        :param message_attribute_names: The name of the message attribute, where *N* is the index.
        :param max_number_of_messages: The maximum number of messages to return.
        :param visibility_timeout: The duration (in seconds) that the received messages are hidden from
        subsequent retrieve requests after being retrieved by a
        ``ReceiveMessage`` request.
        :param wait_time_seconds: The duration (in seconds) for which the call waits for a message to
        arrive in the queue before returning.
        :param receive_request_attempt_id: This parameter applies only to FIFO (first-in-first-out) queues.
        :returns: ReceiveMessageResult
        :raises OverLimit:
        """
        raise NotImplementedError

    @handler("RemovePermission")
    def remove_permission(self, context: RequestContext, queue_url: String, label: String) -> None:
        """Revokes any permissions in the queue policy that matches the specified
        ``Label`` parameter.

        -  Only the owner of a queue can remove permissions from it.

        -  Cross-account permissions don't apply to this action. For more
           information, see `Grant cross-account permissions to a role and a
           user
           name <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-customer-managed-policy-examples.html#grant-cross-account-permissions-to-role-and-user-name>`__
           in the *Amazon SQS Developer Guide*.

        -  To remove the ability to change queue permissions, you must deny
           permission to the ``AddPermission``, ``RemovePermission``, and
           ``SetQueueAttributes`` actions in your IAM policy.

        :param queue_url: The URL of the Amazon SQS queue from which permissions are removed.
        :param label: The identification of the permission to remove.
        """
        raise NotImplementedError

    @handler("SendMessage")
    def send_message(
        self,
        context: RequestContext,
        queue_url: String,
        message_body: String,
        delay_seconds: Integer = None,
        message_attributes: MessageBodyAttributeMap = None,
        message_system_attributes: MessageBodySystemAttributeMap = None,
        message_deduplication_id: String = None,
        message_group_id: String = None,
    ) -> SendMessageResult:
        r"""Delivers a message to the specified queue.

        A message can include only XML, JSON, and unformatted text. The
        following Unicode characters are allowed:

        ``#x9`` \| ``#xA`` \| ``#xD`` \| ``#x20`` to ``#xD7FF`` \| ``#xE000`` to
        ``#xFFFD`` \| ``#x10000`` to ``#x10FFFF``

        Any characters not included in this list will be rejected. For more
        information, see the `W3C specification for
        characters <http://www.w3.org/TR/REC-xml/#charsets>`__.

        :param queue_url: The URL of the Amazon SQS queue to which a message is sent.
        :param message_body: The message to send.
        :param delay_seconds: The length of time, in seconds, for which to delay a specific message.
        :param message_attributes: Each message attribute consists of a ``Name``, ``Type``, and ``Value``.
        :param message_system_attributes: The message system attribute to send.
        :param message_deduplication_id: This parameter applies only to FIFO (first-in-first-out) queues.
        :param message_group_id: This parameter applies only to FIFO (first-in-first-out) queues.
        :returns: SendMessageResult
        :raises InvalidMessageContents:
        :raises UnsupportedOperation:
        """
        raise NotImplementedError

    @handler("SendMessageBatch")
    def send_message_batch(
        self,
        context: RequestContext,
        queue_url: String,
        entries: SendMessageBatchRequestEntryList,
    ) -> SendMessageBatchResult:
        r"""Delivers up to ten messages to the specified queue. This is a batch
        version of ``SendMessage.`` For a FIFO queue, multiple messages within a
        single batch are enqueued in the order they are sent.

        The result of sending each message is reported individually in the
        response. Because the batch request can result in a combination of
        successful and unsuccessful actions, you should check for batch errors
        even when the call returns an HTTP status code of ``200``.

        The maximum allowed individual message size and the maximum total
        payload size (the sum of the individual lengths of all of the batched
        messages) are both 256 KB (262,144 bytes).

        A message can include only XML, JSON, and unformatted text. The
        following Unicode characters are allowed:

        ``#x9`` \| ``#xA`` \| ``#xD`` \| ``#x20`` to ``#xD7FF`` \| ``#xE000`` to
        ``#xFFFD`` \| ``#x10000`` to ``#x10FFFF``

        Any characters not included in this list will be rejected. For more
        information, see the `W3C specification for
        characters <http://www.w3.org/TR/REC-xml/#charsets>`__.

        If you don't specify the ``DelaySeconds`` parameter for an entry, Amazon
        SQS uses the default value for the queue.

        Some actions take lists of parameters. These lists are specified using
        the ``param.n`` notation. Values of ``n`` are integers starting from 1.
        For example, a parameter list with two elements looks like this:

        ``&AttributeName.1=first``

        ``&AttributeName.2=second``

        :param queue_url: The URL of the Amazon SQS queue to which batched messages are sent.
        :param entries: A list of ``SendMessageBatchRequestEntry`` items.
        :returns: SendMessageBatchResult
        :raises TooManyEntriesInBatchRequest:
        :raises EmptyBatchRequest:
        :raises BatchEntryIdsNotDistinct:
        :raises BatchRequestTooLong:
        :raises InvalidBatchEntryId:
        :raises UnsupportedOperation:
        """
        raise NotImplementedError

    @handler("SetQueueAttributes")
    def set_queue_attributes(
        self, context: RequestContext, queue_url: String, attributes: QueueAttributeMap
    ) -> None:
        """Sets the value of one or more queue attributes. When you change a
        queue's attributes, the change can take up to 60 seconds for most of the
        attributes to propagate throughout the Amazon SQS system. Changes made
        to the ``MessageRetentionPeriod`` attribute can take up to 15 minutes.

        -  In the future, new attributes might be added. If you write code that
           calls this action, we recommend that you structure your code so that
           it can handle new attributes gracefully.

        -  Cross-account permissions don't apply to this action. For more
           information, see `Grant cross-account permissions to a role and a
           user
           name <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-customer-managed-policy-examples.html#grant-cross-account-permissions-to-role-and-user-name>`__
           in the *Amazon SQS Developer Guide*.

        -  To remove the ability to change queue permissions, you must deny
           permission to the ``AddPermission``, ``RemovePermission``, and
           ``SetQueueAttributes`` actions in your IAM policy.

        :param queue_url: The URL of the Amazon SQS queue whose attributes are set.
        :param attributes: A map of attributes to set.
        :raises InvalidAttributeName:
        """
        raise NotImplementedError

    @handler("TagQueue")
    def tag_queue(self, context: RequestContext, queue_url: String, tags: TagMap) -> None:
        """Add cost allocation tags to the specified Amazon SQS queue. For an
        overview, see `Tagging Your Amazon SQS
        Queues <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-queue-tags.html>`__
        in the *Amazon SQS Developer Guide*.

        When you use queue tags, keep the following guidelines in mind:

        -  Adding more than 50 tags to a queue isn't recommended.

        -  Tags don't have any semantic meaning. Amazon SQS interprets tags as
           character strings.

        -  Tags are case-sensitive.

        -  A new tag with a key identical to that of an existing tag overwrites
           the existing tag.

        For a full list of tag restrictions, see `Quotas related to
        queues <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-limits.html#limits-queues>`__
        in the *Amazon SQS Developer Guide*.

        Cross-account permissions don't apply to this action. For more
        information, see `Grant cross-account permissions to a role and a user
        name <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-customer-managed-policy-examples.html#grant-cross-account-permissions-to-role-and-user-name>`__
        in the *Amazon SQS Developer Guide*.

        :param queue_url: The URL of the queue.
        :param tags: The list of tags to be added to the specified queue.
        """
        raise NotImplementedError

    @handler("UntagQueue")
    def untag_queue(self, context: RequestContext, queue_url: String, tag_keys: TagKeyList) -> None:
        """Remove cost allocation tags from the specified Amazon SQS queue. For an
        overview, see `Tagging Your Amazon SQS
        Queues <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-queue-tags.html>`__
        in the *Amazon SQS Developer Guide*.

        Cross-account permissions don't apply to this action. For more
        information, see `Grant cross-account permissions to a role and a user
        name <https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-customer-managed-policy-examples.html#grant-cross-account-permissions-to-role-and-user-name>`__
        in the *Amazon SQS Developer Guide*.

        :param queue_url: The URL of the queue.
        :param tag_keys: The list of tags to be removed from the specified queue.
        """
        raise NotImplementedError



import json
import os
from typing import Callable

from localstack.config import dirs
from localstack.utils import common

# URL to "cfn-response" module which is required in some CF Lambdas
from localstack.utils.common import select_attributes, short_uid

CFN_RESPONSE_MODULE_URL = (
    "https://raw.githubusercontent.com/LukeMizuhashi/cfn-response/master/index.js"
)

# placeholders
PLACEHOLDER_RESOURCE_NAME = "__resource_name__"
PLACEHOLDER_AWS_NO_VALUE = "__aws_no_value__"


def dump_json_params(param_func=None, *param_names):
    def replace(params, **kwargs):
        result = param_func(params, **kwargs) if param_func else params
        for name in param_names:
            if isinstance(result.get(name), (dict, list)):
                # Fix for https://github.com/localstack/localstack/issues/2022
                # Convert any date instances to date strings, etc, Version: "2012-10-17"
                param_value = common.json_safe(result[name])
                result[name] = json.dumps(param_value)
        return result

    return replace


def param_defaults(param_func, defaults):
    def replace(params, **kwargs):
        result = param_func(params, **kwargs)
        for key, value in defaults.items():
            if result.get(key) in ["", None]:
                result[key] = value
        return result

    return replace


def remove_none_values(params):
    """Remove None values and AWS::NoValue placeholders (recursively) in the given object."""

    def remove_nones(o, **kwargs):
        if isinstance(o, dict):
            for k, v in dict(o).items():
                if v in [None, PLACEHOLDER_AWS_NO_VALUE]:
                    o.pop(k)
        if isinstance(o, list):
            common.run_safe(o.remove, None)
            common.run_safe(o.remove, PLACEHOLDER_AWS_NO_VALUE)
        return o

    result = common.recurse_object(params, remove_nones)
    return result


def params_list_to_dict(param_name, key_attr_name="Key", value_attr_name="Value"):
    def do_replace(params, **kwargs):
        result = {}
        for entry in params.get(param_name, []):
            key = entry[key_attr_name]
            value = entry[value_attr_name]
            result[key] = value
        return result

    return do_replace


def lambda_keys_to_lower(key=None):
    return lambda params, **kwargs: common.keys_to_lower(params.get(key) if key else params)


def merge_parameters(func1, func2):
    return lambda params, **kwargs: common.merge_dicts(
        func1(params, **kwargs), func2(params, **kwargs)
    )


def str_or_none(o):
    return o if o is None else json.dumps(o) if isinstance(o, (dict, list)) else str(o)


def params_dict_to_list(param_name, key_attr_name="Key", value_attr_name="Value", wrapper=None):
    def do_replace(params, **kwargs):
        result = []
        for key, value in params.get(param_name, {}).items():
            result.append({key_attr_name: key, value_attr_name: value})
        if wrapper:
            result = {wrapper: result}
        return result

    return do_replace


def params_select_attributes(*attrs):
    def do_select(params, **kwargs):
        result = {}
        for attr in attrs:
            if params.get(attr) is not None:
                result[attr] = str_or_none(params.get(attr))
        return result

    return do_select


def param_json_to_str(name):
    def _convert(params, **kwargs):
        result = params.get(name)
        if result:
            result = json.dumps(result)
        return result

    return _convert


def get_cfn_response_mod_file():
    cfn_response_tmp_file = os.path.join(dirs.static_libs, "lambda.cfn-response.js")
    if not os.path.exists(cfn_response_tmp_file):
        common.download(CFN_RESPONSE_MODULE_URL, cfn_response_tmp_file)
    return cfn_response_tmp_file


def lambda_select_params(*selected):
    # TODO: remove and merge with function below
    return select_parameters(*selected)


def select_parameters(*param_names):
    return lambda params, **kwargs: select_attributes(params, param_names)


def is_none_or_empty_value(value):
    return not value or value == PLACEHOLDER_AWS_NO_VALUE


def generate_default_name(stack_name: str, logical_resource_id: str):
    random_id_part = short_uid()
    resource_id_part = logical_resource_id[:24]
    stack_name_part = stack_name[: 63 - 2 - (len(random_id_part) + len(resource_id_part))]
    return f"{stack_name_part}-{resource_id_part}-{random_id_part}"


def generate_default_name_without_stack(logical_resource_id: str):
    random_id_part = short_uid()
    resource_id_part = logical_resource_id[: 63 - 1 - len(random_id_part)]
    return f"{resource_id_part}-{random_id_part}"


def pre_create_default_name(key: str) -> Callable[[str, dict, str, dict, str], None]:
    def _pre_create_default_name(
        resource_id: str, resources: dict, resource_type: str, func: dict, stack_name: str
    ):
        resource = resources[resource_id]
        props = resource["Properties"]
        if not props.get(key):
            props[key] = generate_default_name(stack_name, resource_id)

    return _pre_create_default_name



"""
This module provides tools to call moto using moto and botocore internals without going through the moto HTTP server.
"""
import sys
from functools import lru_cache
from typing import Any, Callable, Mapping, Optional, Tuple, Union
from urllib.parse import urlsplit

from botocore.awsrequest import AWSPreparedRequest
from botocore.parsers import create_parser
from moto.backends import get_backend as get_moto_backend
from moto.core.utils import BackendDict
from moto.server import RegexConverter
from werkzeug.datastructures import Headers
from werkzeug.routing import Map, Rule

from localstack import __version__ as localstack_version
from localstack import config
from localstack.aws.api import (
    CommonServiceException,
    HttpRequest,
    HttpResponse,
    RequestContext,
    ServiceResponse,
)
from localstack.aws.api.core import ServiceRequest, ServiceRequestHandler
from localstack.aws.skeleton import DispatchTable, create_dispatch_table
from localstack.aws.spec import load_service
from localstack.utils.aws import aws_stack
from localstack.utils.common import to_bytes, to_str

MotoResponse = Tuple[int, dict, Union[str, bytes]]
MotoDispatcher = Callable[[HttpRequest, str, dict], MotoResponse]

user_agent = f"Localstack/{localstack_version} Python/{sys.version.split(' ')[0]}"


def call_moto(context: RequestContext) -> ServiceResponse:
    """
    Call moto with the given request context and receive a parsed ServiceResponse.

    :param context: the request context
    :return: a serialized AWS ServiceResponse (same as boto3 would return)
    """
    status, headers, content = dispatch_to_moto(context)

    operation_model = context.operation
    response_dict = {  # this is what botocore.endpoint.convert_to_response_dict normally does
        "headers": headers,
        "status_code": status,
        "body": to_bytes(content),
        "context": {
            "operation_name": operation_model.name,
        },
    }

    parser = create_parser(context.service.protocol)
    response = parser.parse(response_dict, operation_model.output_shape)

    if status >= 301:
        error = response["Error"]
        raise CommonServiceException(
            code=error.get("Code", "UnknownError"),
            status_code=status,
            message=error.get("Message", ""),
        )

    return response


def call_moto_with_request(
    context: RequestContext, service_request: Union[Mapping[str, Any], ServiceRequest]
) -> ServiceResponse:
    """
    Like `call_moto`, but you can pass a modified version of the service request before calling moto. The caveat is
    that a new HTTP request has to be created. The service_request is serialized into a new RequestContext object,
    and headers from the old request are merged into the new one.

    :param context: the original request context
    :param service_request: the dictionary containing the service request parameters
    :return: a serialized AWS ServiceResponse (same as boto3 would return)
    """
    local_context = create_aws_request_context(
        service_name=context.service.service_name,
        action=context.operation.name,
        parameters=service_request,
        region=context.region,
    )

    local_context.request.headers.extend(context.request.headers)

    return call_moto(local_context)


def proxy_moto(context: RequestContext) -> HttpResponse:
    """
    Similar to ``call``, only that ``proxy`` does not parse the HTTP response into a ServiceResponse, but instead
    returns directly the HTTP response. This can be useful to pass through moto's response directly to the client.

    :param context: the request context
    :return: the HttpResponse from moto
    """
    status, headers, content = dispatch_to_moto(context)

    return HttpResponse(response=content, status=status, headers=headers)


def MotoFallbackDispatcher(provider: object) -> DispatchTable:
    """
    Wraps a provider with a moto fallthrough mechanism. It does by creating a new DispatchTable from the original
    provider, and wrapping each method with a fallthrough method that calls ``request`` if the original provider
    raises a ``NotImplementedError``.

    :param provider: the ASF provider
    :return: a modified DispatchTable
    """
    table = create_dispatch_table(provider)

    for op, fn in table.items():
        table[op] = _wrap_with_fallthrough(fn)

    return table


def _wrap_with_fallthrough(handler: ServiceRequestHandler) -> ServiceRequestHandler:
    def _call(context, req) -> ServiceResponse:
        try:
            # handler will typically be an ASF provider method, and in case it hasn't been implemented, we try to
            # fall through to moto
            return handler(context, req)
        except NotImplementedError:
            return proxy_moto(context)

    return _call


def dispatch_to_moto(context: RequestContext) -> MotoResponse:
    """
    Internal method to dispatch the request to moto without changing moto's dispatcher output.
    :param context: the request context
    :return: the response from moto
    """
    service = context.service
    request = context.request

    # hack to avoid call to request.form (see moto's BaseResponse.dispatch)
    request.body = request.data

    # this is where we skip the HTTP roundtrip between the moto server and the boto client
    dispatch = get_dispatcher(service.service_name, request.path)

    return dispatch(request, request.url, request.headers)


def get_dispatcher(service: str, path: str) -> MotoDispatcher:
    url_map = get_moto_routing_table(service)

    if len(url_map._rules) == 1:
        # in most cases, there will only be one dispatch method in the list of urls, so no need to do matching
        rule = next(url_map.iter_rules())
        return rule.endpoint

    matcher = url_map.bind(config.LOCALSTACK_HOSTNAME)
    endpoint, _ = matcher.match(path_info=path)
    return endpoint


@lru_cache()
def get_moto_routing_table(service: str) -> Map:
    """Cached version of load_moto_routing_table."""
    return load_moto_routing_table(service)


def load_moto_routing_table(service: str) -> Map:
    """
    Creates from moto service url_paths a werkzeug URL rule map that can be used to locate moto methods to dispatch
    requests to.

    :param service: the service to get the map for.
    :return: a new Map object
    """
    # code from moto.server.create_backend_app
    backend_dict: BackendDict = get_moto_backend(service)
    if "us-east-1" in backend_dict:
        backend = backend_dict["us-east-1"]
    else:
        backend = backend_dict["global"]

    url_map = Map()
    url_map.converters["regex"] = RegexConverter

    for url_path, handler in backend.flask_paths.items():
        # endpoints are annotated as string in werkzeug, but don't have to be
        url_map.add(Rule(url_path, endpoint=handler))

    return url_map


def create_aws_request_context(
    service_name: str,
    action: str,
    parameters: Mapping[str, Any] = None,
    region: str = None,
    endpoint_url: Optional[str] = None,
) -> RequestContext:
    """
    This is a stripped-down version of what the botocore client does to perform an HTTP request from a client call. A
    client call looks something like this: boto3.client("sqs").create_queue(QueueName="myqueue"), which will be
    serialized into an HTTP request. This method does the same, without performing the actual request, and with a
    more low-level interface. An equivalent call would be

         create_aws_request_context("sqs", "CreateQueue", {"QueueName": "myqueue"})

    :param service_name: the AWS service
    :param action: the action to invoke
    :param parameters: the invocation parameters
    :param region: the region name (default is us-east-1)
    :param endpoint_url: the endpoint to call (defaults to localstack)
    :return: a RequestContext object that describes this request
    """
    if parameters is None:
        parameters = {}
    if region is None:
        region = config.AWS_REGION_US_EAST_1

    service = load_service(service_name)
    operation = service.operation_model(action)

    # we re-use botocore internals here to serialize the HTTP request, but don't send it
    client = aws_stack.connect_to_service(
        service_name, endpoint_url=endpoint_url, region_name=region
    )
    request_context = {
        "client_region": region,
        "has_streaming_input": operation.has_streaming_input,
        "auth_type": operation.auth_type,
    }
    request_dict = client._convert_to_request_dict(parameters, operation, context=request_context)
    aws_request = client._endpoint.create_request(request_dict, operation)

    context = RequestContext()
    context.service = service
    context.operation = operation
    context.region = region
    context.request = create_http_request(aws_request)

    return context


def create_http_request(aws_request: AWSPreparedRequest) -> HttpRequest:
    # create HttpRequest from AWSRequest
    split_url = urlsplit(aws_request.url)
    host = split_url.netloc.split(":")
    if len(host) == 1:
        server = (to_str(host[0]), None)
    elif len(host) == 2:
        server = (to_str(host[0]), int(host[1]))
    else:
        raise ValueError

    # prepare the RequestContext
    headers = Headers()
    for k, v in aws_request.headers.items():
        headers[k] = v

    return HttpRequest(
        method=aws_request.method,
        path=split_url.path,
        query_string=split_url.query,
        headers=headers,
        body=aws_request.body,
        server=server,
    )



from localstack.services.cloudformation.deployment_utils import generate_default_name
from localstack.services.cloudformation.service_models import GenericBaseModel
from localstack.utils.aws import aws_stack


class CloudWatchAlarm(GenericBaseModel):
    @staticmethod
    def cloudformation_type():
        return "AWS::CloudWatch::Alarm"

    def get_physical_resource_id(self, attribute=None, **kwargs):
        return self.props.get("AlarmName")

    def _response_name(self):
        return "MetricAlarms"

    @classmethod
    def _create_function_name(self):
        return "put_metric_alarm"

    def fetch_state(self, stack_name, resources):
        client = aws_stack.connect_to_service("cloudwatch")
        alarm_name = self.resolve_refs_recursively(stack_name, self.props["AlarmName"], resources)
        result = client.describe_alarms(AlarmNames=[alarm_name]).get(self._response_name(), [])
        return (result or [None])[0]

    @staticmethod
    def add_defaults(resource, stack_name: str):
        role_name = resource.get("Properties", {}).get("AlarmName")
        if not role_name:
            resource["Properties"]["AlarmName"] = generate_default_name(
                stack_name, resource["LogicalResourceId"]
            )

    @classmethod
    def get_deploy_templates(cls):
        def get_delete_params(params, **kwargs):
            return {"AlarmNames": [params["AlarmName"]]}

        return {
            "create": {"function": cls._create_function_name()},
            "delete": {"function": "delete_alarms", "parameters": get_delete_params},
        }


class CloudWatchCompositeAlarm(CloudWatchAlarm):
    @staticmethod
    def cloudformation_type():
        return "AWS::CloudWatch::CompositeAlarm"

    def _response_name(self):
        return "CompositeAlarms"

    @classmethod
    def _create_function_name(self):
        return "put_composite_alarm"



import logging

from localstack import config
from localstack.utils.container_utils.container_client import ContainerClient

"""Type alias for a simple version of VolumeBind"""

LOG = logging.getLogger(__name__)


def is_docker_sdk_installed() -> bool:
    try:
        import docker  # noqa: F401

        return True
    except ModuleNotFoundError:
        return False


def create_docker_client() -> ContainerClient:
    if config.LEGACY_DOCKER_CLIENT or not is_docker_sdk_installed():
        from localstack.utils.container_utils.docker_cmd_client import CmdDockerClient

        LOG.debug(
            "Using CmdDockerClient. LEGACY_DOCKER_CLIENT: %s, SDK installed: %s",
            config.LEGACY_DOCKER_CLIENT,
            is_docker_sdk_installed(),
        )

        return CmdDockerClient()
    else:
        from localstack.utils.container_utils.docker_sdk_client import SdkDockerClient

        LOG.debug(
            "Using SdkDockerClient. LEGACY_DOCKER_CLIENT: %s, SDK installed: %s",
            config.LEGACY_DOCKER_CLIENT,
            is_docker_sdk_installed(),
        )

        return SdkDockerClient()


DOCKER_CLIENT: ContainerClient = create_docker_client()



import logging
import types
from html import escape
from typing import Optional

from moto.core.utils import camelcase_to_underscores
from moto.sqs import responses as sqs_responses
from moto.sqs.exceptions import QueueDoesNotExist
from moto.sqs.models import Queue

from localstack import config
from localstack.services.infra import start_moto_server, start_proxy_for_service
from localstack.services.install import SQS_BACKEND_IMPL
from localstack.services.sqs.elasticmq import ElasticMQSerer
from localstack.utils.aws import aws_stack
from localstack.utils.common import get_free_tcp_port, to_str
from localstack.utils.patch import patch
from localstack.utils.serving import Server

LOG = logging.getLogger(__name__)

# backend port (configured at startup)
PORT_SQS_BACKEND = None

# max heap size allocated for the Java process
MAX_HEAP_SIZE = "256m"


# server singleton
_server: Optional[Server] = None


def check_sqs(expect_shutdown=False, print_error=False):
    out = None

    try:
        if not expect_shutdown:
            assert _server, "server has not been started yet"
            assert _server.wait_is_up(5), "gave up waiting for server"

        # check SQS
        endpoint_url = _server.url
        out = aws_stack.connect_to_service(
            service_name="sqs", endpoint_url=endpoint_url
        ).list_queues()
    except Exception:
        if print_error:
            LOG.exception("SQS health check failed")
    if expect_shutdown:
        assert out is None
    else:
        assert out is not None
        assert out.get("ResponseMetadata", {}).get("HTTPStatusCode") == 200


def start_sqs(*args, **kwargs):
    global _server, PORT_SQS_BACKEND

    if _server:
        return _server

    if SQS_BACKEND_IMPL == "elasticmq":
        _server = start_sqs_elasticmq(*args, **kwargs)
    else:
        _server = start_sqs_moto(*args, **kwargs)

    PORT_SQS_BACKEND = _server.port

    return _server


def patch_moto():
    # patch add_message to disable event source mappings in moto
    @patch(Queue.add_message)
    def add_message(fn, self, *args, **kwargs):
        mappings = self.lambda_event_source_mappings
        try:
            # temporarily set mappings to empty dict, to prevent moto from consuming messages from the queue
            self.lambda_event_source_mappings = {}
            return fn(self, *args, **kwargs)
        finally:
            self.lambda_event_source_mappings = mappings

    @patch(Queue._set_attributes)
    def _set_attributes(fn, self, attributes, now=None):
        fn(self, attributes, now)

        integer_fields = ["ReceiveMessageWaitTimeSeconds"]

        for key in integer_fields:
            attribute = camelcase_to_underscores(key)
            setattr(self, attribute, int(getattr(self, attribute, 0)))

    # pass additional globals (e.g., escaping methods) to template render method
    @patch(sqs_responses.SQSResponse.response_template)
    def response_template(fn, self, template_str, *args, **kwargs):
        template = fn(self, template_str, *args, **kwargs)

        def _escape(val):
            try:
                return val and escape(to_str(val))
            except Exception:
                return val

        def render(self, *args, **kwargs):
            return render_orig(*args, _escape=_escape, **kwargs)

        if not hasattr(template, "__patched"):
            render_orig = template.render
            template.render = types.MethodType(render, template)
            template.__patched = True
        return template

    # escape message responses to allow for special characters like "<"
    sqs_responses.RECEIVE_MESSAGE_RESPONSE = sqs_responses.RECEIVE_MESSAGE_RESPONSE.replace(
        "<StringValue><![CDATA[{{ value.string_value }}]]></StringValue>",
        "<StringValue>{{ _escape(value.string_value) }}</StringValue>",
    )

    # Fix issue with trailing slash
    # https://github.com/localstack/localstack/issues/2874
    @patch(sqs_responses.SQSResponse._get_queue_name, False)
    def sqs_responses_get_queue_name(self):
        try:
            queue_url = self.querystring.get("QueueUrl")[0]
            queue_name_data = queue_url.split("/")[4:]
            queue_name_data = [queue_attr for queue_attr in queue_name_data if queue_attr]
            queue_name = "/".join(queue_name_data)
        except TypeError:
            # Fallback to reading from the URL
            queue_name = self.path.split("/")[2]

        if not queue_name:
            raise QueueDoesNotExist()

        return queue_name


def start_sqs_moto(port=None, asynchronous=False, update_listener=None) -> Server:
    from localstack.services import motoserver

    port = port or config.service_port("sqs")
    patch_moto()
    start_moto_server(
        "sqs",
        port,
        name="SQS",
        asynchronous=asynchronous,
        update_listener=update_listener,
    )

    return motoserver.get_moto_server()


def start_sqs_elasticmq(port=None, asynchronous=False, update_listener=None) -> Server:
    server = ElasticMQSerer(get_free_tcp_port())
    server.start()
    start_proxy_for_service("sqs", port, server.port, update_listener)
    LOG.debug("waiting for elasticmq server to start...")
    if not server.wait_is_up(120):
        LOG.debug("gave up waiting for elasticmq server after 120 seconds")
    return server



import base64
import json
import re
from urllib.parse import quote_plus, unquote_plus

from localstack import config
from localstack.utils.common import (
    extract_jsonpath,
    is_number,
    json_safe,
    recurse_object,
    short_uid,
)
from localstack.utils.generic.number_utils import to_number


class VelocityInput(object):
    """Simple class to mimick the behavior of variable '$input' in AWS API Gateway integration velocity templates.
    See: http://docs.aws.amazon.com/apigateway/latest/developerguide/api-gateway-mapping-template-reference.html"""

    def __init__(self, value):
        self.value = value

    def path(self, path):
        if not self.value:
            return {}
        value = self.value if isinstance(self.value, dict) else json.loads(self.value)
        return extract_jsonpath(value, path)

    def json(self, path):
        path = path or "$"
        matching = self.path(path)
        if isinstance(matching, (list, dict)):
            matching = json_safe(matching)
        return json.dumps(matching)

    def __getattr__(self, name):
        return self.value.get(name)

    def __repr__(self):
        return "$input"


class VelocityUtil(object):
    """Simple class to mimick the behavior of variable '$util' in AWS API Gateway integration velocity templates.
    See: http://docs.aws.amazon.com/apigateway/latest/developerguide/api-gateway-mapping-template-reference.html"""

    def base64Encode(self, s):
        if not isinstance(s, str):
            s = json.dumps(s)
        encoded_str = s.encode(config.DEFAULT_ENCODING)
        encoded_b64_str = base64.b64encode(encoded_str)
        return encoded_b64_str.decode(config.DEFAULT_ENCODING)

    def base64Decode(self, s):
        if not isinstance(s, str):
            s = json.dumps(s)
        return base64.b64decode(s)

    def toJson(self, obj):
        return obj and json.dumps(obj)

    def urlEncode(self, s):
        return quote_plus(s)

    def urlDecode(self, s):
        return unquote_plus(s)

    def escapeJavaScript(self, s):
        try:
            return json.dumps(json.loads(s))
        except Exception:
            primitive_types = (str, int, bool, float, type(None))
            s = s if isinstance(s, primitive_types) else str(s)
        if str(s).strip() in ["true", "false"]:
            s = bool(s)
        elif s not in [True, False] and is_number(s):
            s = to_number(s)
        return json.dumps(s)


def render_velocity_template(template, context, variables=None, as_json=False):
    if variables is None:
        variables = {}
    import airspeed

    if not template:
        return template

    # fix "#set" commands
    template = re.sub(r"(^|\n)#\s+set(.*)", r"\1#set\2", template, re.MULTILINE)

    # enable syntax like "test#${foo.bar}"
    empty_placeholder = " __pLaCe-HoLdEr__ "
    template = re.sub(
        r"([^\s]+)#\$({)?(.*)",
        r"\1#%s$\2\3" % empty_placeholder,
        template,
        re.MULTILINE,
    )

    # add extensions for common string functions below

    class ExtendedString(str):
        def trim(self, *args, **kwargs):
            return ExtendedString(self.strip(*args, **kwargs))

        def toLowerCase(self, *args, **kwargs):
            return ExtendedString(self.lower(*args, **kwargs))

        def toUpperCase(self, *args, **kwargs):
            return ExtendedString(self.upper(*args, **kwargs))

    def apply(obj, **kwargs):
        if isinstance(obj, dict):
            for k, v in obj.items():
                if isinstance(v, str):
                    obj[k] = ExtendedString(v)
        return obj

    # loop through the variables and enable certain additional util functions (e.g., string utils)
    variables = variables or {}
    recurse_object(variables, apply)

    # prepare and render template
    context_var = variables.get("context") or {}
    context_var.setdefault("requestId", short_uid())
    t = airspeed.Template(template)
    var_map = {
        "input": VelocityInput(context),
        "util": VelocityUtil(),
        "context": context_var,
    }
    var_map.update(variables or {})
    replaced = t.merge(var_map)

    # revert temporary changes from the fixes above
    replaced = replaced.replace(empty_placeholder, "")

    if as_json:
        replaced = json.loads(replaced)
    return replaced
